from decimal import Decimal
from typing import Any, Callable, Dict, List, Union

from returns.result import Result

ValueTypes = (str, int, float, bool, Decimal)

Value = Union[str, int, float, bool, Decimal]
ValueDict = Dict[str, Value]
ValueDictList = List[ValueDict]
NewValue = Union[dict, list, Value]

MapValue = Value
MapAttribute = Union[Value, ValueDict, ValueDictList]
MapObject = Dict[str, MapAttribute]
MapResult = Union[MapObject, List[MapObject]]

# Callables
CallableValidateAndMarshall = Callable[[dict], Result[dict, Exception]]
CallablePreProcess = Callable[[Any], Result[dict, Exception]]
CallableOutput = Callable[[dict], Result[Any, Exception]]
CallableLogFunction = Callable[[Any], None]

CallableMapFunction = Callable[
    [
        Union[Dict[str, Any], List[Any]],
        Dict[str, Any],
    ],
    Result[MapResult, Exception],
]



import datetime
import decimal
import re
from typing import Callable, Optional

from attr import dataclass
from returns.pipeline import flow
from returns.pointfree import alt, bind, map_, rescue
from returns.result import Failure, ResultE, safe
from typing_extensions import final

from piri.constants import (
    COMMA,
    DATE,
    DECIMAL,
    EMPTY,
    INTEGER,
    INTEGER_CONTAINING_DECIMALS,
    PERIOD,
)
from piri.valuetypes import MapValue


@safe
def get_casting_function(cast_to: str) -> Callable:
    """Return casting function depending on name."""
    if cast_to == INTEGER:
        return CastToInteger()

    elif cast_to == DECIMAL:
        return CastToDecimal()

    elif cast_to == DATE:
        return CastToDate()

    raise NotImplementedError(
        'Unsupported cast to value ({0})'.format(cast_to),
    )


@final
@dataclass(frozen=True, slots=True)
class CastToDecimal(object):
    """Cast input to decimal."""

    _decimal_pattern = re.compile(r'^([0-9]|-|\.|,)+$')
    _decimal_with_period_after_commas = re.compile(r'^-?(\d+\,)*\d+\.\d+$')
    _decimal_with_comma_after_periods = re.compile(r'^-?(\d+\.)*\d+\,\d+$')

    @safe
    def __call__(
        self,
        value_to_cast: MapValue,
        original_format: Optional[str] = None,
    ) -> decimal.Decimal:
        """Make this object callable."""
        the_value = str(value_to_cast).replace(' ', EMPTY)

        if not self._decimal_pattern.match(the_value):
            raise ValueError(
                "Illegal characters in value '{0}'".format(the_value),
            )

        if original_format == INTEGER_CONTAINING_DECIMALS:
            return decimal.Decimal(the_value) / 100

        # ie 1234567,89 only comma as decimal separator
        if the_value.count(COMMA) == 1 and not the_value.count(PERIOD):
            return decimal.Decimal(the_value.replace(COMMA, PERIOD))

        # ie 1,234,567.89 many commas followed by one period
        if self._decimal_with_period_after_commas.match(the_value):
            return decimal.Decimal(the_value.replace(COMMA, EMPTY))

        # ie 1.234.567,89 many periods followed by one comma
        if self._decimal_with_comma_after_periods.match(the_value):
            return decimal.Decimal(
                the_value.replace(PERIOD, EMPTY).replace(COMMA, PERIOD),
            )

        return decimal.Decimal(the_value)


@final
@dataclass(frozen=True, slots=True)
class CastToInteger(object):
    """Cast input to integer."""

    _cast_to_decimal = CastToDecimal()

    def __call__(
        self,
        value_to_cast: MapValue,
        original_format: Optional[str] = None,
    ) -> ResultE[int]:
        """Make this object callable."""
        return flow(
            value_to_cast,
            self._cast_to_decimal,
            map_(quantize_decimal),
            map_(int),
        )


def quantize_decimal(number: decimal.Decimal) -> decimal.Decimal:
    """Quantize a decimal to whole number."""
    return number.quantize(decimal.Decimal('1.'))


@final
@dataclass(frozen=True, slots=True)
class CastToDate(object):
    """Cast input to date."""

    _yymmdd_pattern = re.compile(r'^yy[^\w]?mm[^\w]?dd$')
    _ddmmyy_pattern = re.compile(r'^dd[^\w]?mm[^\w]?yy$')
    _mmddyy_pattern = re.compile(r'^mm[^\w]?dd[^\w]?yy$')

    _mmddyyyy_pattern = re.compile(r'^mm[^\w]?dd[^\w]?yyyy$')
    _ddmmyyyy_pattern = re.compile(r'^dd[^\w]?mm[^\w]?yyyy$')
    _yyyymmdd_pattern = re.compile(r'^yyyy[^\w]?mm[^\w]?dd$')
    _iso_pattern = re.compile(r'^\d{4}-\d{2}-\d{2}')

    _error_message = 'Unable to cast ({value}) to ISO date. Exc({failure})'

    # @pipeline(ResultE)
    def __call__(
        self,
        value_to_cast: MapValue,
        original_format: str,
    ) -> ResultE[MapValue]:
        r"""
        purpose: Convert string date into ISO format date.

        :input value_to_cast: the string to be converted to a date
        :input original_format: a string describing the format the data string
                             is before convertion.
        :raises ValueError:  when not able to convert or value_to_cast
                             does not match original_format.
        regex_tips  ^ = start of string
                    $ = end of string
                    [^[0-9]] = matches anything not [0-9]
                    [^\d] = matches anything not digits
                    \d = matches digits
                    {n} = matches previous match n amount of times
                    () = grouper, groups up stuff for use in replace.
        """
        date_value: str = str(value_to_cast)
        return flow(
            date_value,
            self._value_is_iso,
            rescue(  # type: ignore
                lambda _: self._cast_with_millennia(
                    date_value,
                    original_format=original_format,
                ),
            ),
            rescue(  # type: ignore
                lambda _: self._cast_with_no_millennia(
                    date_value,
                    original_format=original_format,
                ),
            ),
            bind(self._validate_date),
            alt(  # type: ignore
                lambda failure: ValueError(
                    self._error_message.format(
                        value=date_value,
                        failure=failure,
                    ),
                ),
            ),
        )

    @safe
    def _value_is_iso(self, value_to_cast: str) -> str:
        if self._iso_pattern.match(value_to_cast):
            return value_to_cast
        raise ValueError('Unable to cast to ISO date')

    def _cast_with_millennia(
        self,
        value_to_cast: str,
        original_format: str,
    ) -> ResultE[str]:
        # mm[.]dd[.]yyyy any separator
        if self._mmddyyyy_pattern.match(original_format):
            return self._apply_regex_sub(
                r'(\d{2})[^\w]?(\d{2})[^\w]?(\d{4})',
                r'\3-\1-\2',
                value_to_cast,
            )

        # dd[.]mm[.]yyyy any separator
        if self._ddmmyyyy_pattern.match(original_format):
            return self._apply_regex_sub(
                r'(\d{2})[^\w]?(\d{2})[^\w]?(\d{4})',
                r'\3-\2-\1',
                value_to_cast,
            )

        # yyyy[.]mm[.]dd any separator
        if self._yyyymmdd_pattern.match(original_format):
            return self._apply_regex_sub(
                r'(\d{4})[^\w]?(\d{2})[^\w]?(\d{2})',
                r'\1-\2-\3',
                value_to_cast,
            )

        return Failure(
            ValueError(
                'Unable to case to milennia format: {value}'.format(
                    value=value_to_cast,
                ),
            ),
        )

    def _cast_with_no_millennia(
        self,
        value_to_cast: str,
        original_format: str,
    ) -> ResultE[str]:
        no_millenia_patterns = {
            self._yymmdd_pattern: (
                r'(\d{2})[^\w]?(\d{2})[^\w]?(\d{2})',
                r'\3\2\1',
            ),
            self._ddmmyy_pattern: (
                r'(\d{2})[^\w]?(\d{2})[^\w]?(\d{2})',
                r'\1\2\3',
            ),
            self._mmddyy_pattern: (
                r'(\d{2})[^\w]?(\d{2})[^\w]?(\d{2})',
                r'\2\1\3',
            ),
        }
        for no_millenia_pattern in no_millenia_patterns:  # noqa: WPS528
            if no_millenia_pattern.match(original_format):
                pattern, arrangement = no_millenia_patterns[no_millenia_pattern]
                return self._apply_regex_sub(
                    pattern, arrangement, value_to_cast,
                ).bind(
                    self._convert_ddmmyy_to_iso_date,
                )

        return Failure(
            ValueError(
                'Unable to cast to no millennia format: {value}'.format(
                    value=value_to_cast,
                ),
            ),
        )

    @safe
    def _validate_date(self, date_string: str) -> str:
        return datetime.date(
            *map(
                int,
                date_string.split('-'),
            ),
        ).isoformat()

    @safe
    def _apply_regex_sub(
        self,
        pattern: str,
        arrangement: str,
        input_string: str,
    ) -> str:
        return re.sub(pattern, arrangement, input_string)

    @safe  # noqa: WPS210, Hard to reduce variables
    def _convert_ddmmyy_to_iso_date(  # noqa: WPS210
        self,
        date_string: str,
    ) -> str:

        in_day = int(date_string[:2])
        in_month = int(date_string[2:4])
        in_year = int(date_string[4:6])

        today = datetime.date.today()
        # think of this 'century' variable as century - 1. as in:
        # 2018 = 2000, 1990 = 1900
        century = (today.year // 100) * 100
        # yy = last two digits of year. 2018 = 18, 1990 = 90
        yy = today.year % 100
        if in_year > yy:
            century -= 100
        year = century + in_year
        date_obj = datetime.date(year, in_month, in_day)

        return date_obj.isoformat()



from typing import Any, Dict, List, Union

from returns.result import Failure, ResultE, Success, safe

from piri.constants import ALIAS, PATH
from piri.valuetypes import MapValue


@safe
def set_value_in_dict(
    new_value: MapValue,
    collection: Dict[str, Any],
    path: List[str],
) -> None:
    """Set value in a dict(pass by ref) by path."""
    if not path:
        raise ValueError('path list empty')

    for key in path[:-1]:
        # this will return a Failure[KeyError] if not found
        collection = collection[key]

    collection[path[-1]] = new_value


@safe
def fetch_data_by_keys(
    collection: Union[Dict[str, Any], List[Any]],
    path: List[Union[str, int]],
) -> MapValue:
    """Find data in collection by following a list of path."""
    if not path:
        raise ValueError('path list empty')

    for key in path:
        # this will return a Failure[KeyError] if not found
        collection = collection[key]  # type: ignore

    return collection  # type: ignore


@safe
def fetch_list_by_keys(
    collection: Dict[str, Any],
    path: List[str],
) -> list:
    """Find data that *must* be a list else it fails.

    Example
        >>> fetch_list_by_keys(
        ...     {'object': {'some_list': ['1']}}, ['object', 'some_list'],
        ... ).unwrap()
        ['1']
    """
    if not path:
        raise ValueError('path list empty')

    for key in path:
        # this will return a Failure[KeyError] if not found
        collection = collection[key]

    if isinstance(collection, list):  # type: ignore
        return collection  # type: ignore

    raise ValueError('Non list data found: ', str(collection))


def iterable_data_handler(raw_data, paths) -> ResultE[list]:
    """Iterate and create all combinations from list of paths."""
    if not paths:
        return Failure(ValueError('No paths'))

    path, rest = paths[0], paths[1:]

    if not rest:
        return create_iterable(raw_data, path)

    my_list: list = []

    for iterable in create_iterable(raw_data, path).unwrap():

        iterable_data_handler(iterable, rest).map(
            my_list.extend,
        )

    return Success(my_list)


def create_iterable(input_data, iterable) -> ResultE[list]:
    """Return set of set of data per entry in list at iterable[path]."""
    return fetch_list_by_keys(
        input_data,
        iterable[PATH],
    ).map(
        lambda collections: [
            {
                **input_data,
                **{iterable[ALIAS]: collection},
            }
            for collection in collections
        ],
    ).fix(
        lambda _: [input_data],
    )



from typing import Any, overload

from attr import dataclass
from returns.result import ResultE, safe
from typing_extensions import Literal, final


@final
@dataclass(frozen=True, slots=True)
class ReadLocalFile(object):
    """Reads local file.

    :param file_path: path to the file to read.
    :type file_path: str

    :parm mode: read mode, r = read string, rb =read bytes.
    :type mode: str

    :return: Success[bytes], Success[str], Failure[Exception]
    :rtype: bytes, str
    """

    @overload
    def __call__(
        self, file_path: str, mode: Literal['r'],
    ) -> ResultE[str]:
        """When 'r' is supplied we return 'str'."""

    @overload  # noqa: WPS440, F811
    def __call__(
        self, file_path: str, mode: Literal['rb'],
    ) -> ResultE[bytes]:
        """When 'rb' is supplied we return 'bytes' instead of a 'str'."""

    @overload  # noqa: WPS440, F811
    def __call__(self, file_path: str, mode: str) -> ResultE[Any]:
        """Any other options might return Any-thing!."""

    @safe  # noqa: WPS440, F811
    def __call__(self, file_path: str, mode: str) -> Any:
        """Open the file and return its contents."""
        with open(file_path, mode) as data_file:
            return data_file.read()



from typing import Any, Dict, List, Union

from returns.curry import partial
from returns.pipeline import flow, is_successful
from returns.pointfree import bind, fix, map_, rescue
from returns.result import ResultE

from piri.collection_handlers import fetch_data_by_keys
from piri.constants import (
    CASTING,
    DEFAULT,
    IF_STATEMENTS,
    MAPPINGS,
    PATH,
    REGEXP,
    SEPARATOR,
    SLICING,
)
from piri.functions import (
    apply_casting,
    apply_default,
    apply_if_statements,
    apply_regexp,
    apply_separator,
    apply_slicing,
)
from piri.valuetypes import MapValue


def handle_mapping(
    collection: Union[Dict[str, Any], List[Any]],
    cfg: Dict[str, Any],
) -> ResultE[MapValue]:
    """Find data in path and apply if statements or default value.

    .. versionadded:: 0.0.1

    :param configuration: :term:`configuration` data to use when mapping
    :type configuration: Dict[str, Any]

    :param collection: The collection of data to find data in
    :type collection: Union[Dict[str, Any], List[Any]]

    :return: Success/Failure containers
    :rtype: GoResult

    configuration expected to look like this:

    .. code-block:: json
        {
            "path": [],
            "if_statementss": [{}, {}],
            "default": 'val'
        }

    Flow description:

    find data from path or None ->
    apply if statements ->
    return default value if Failure else mapped value
    """
    return flow(
        collection,
        partial(fetch_data_by_keys, path=cfg.get(PATH, [])),
        fix(lambda _: None),  # type: ignore
        bind(
            partial(
                apply_regexp, regexp=cfg.get(REGEXP, {}),
            ),
        ),
        fix(lambda _: None),  # type: ignore
        map_(partial(
            apply_slicing, slicing=cfg.get(SLICING, {}),
        )),
        bind(partial(
            apply_if_statements, if_objects=cfg.get(IF_STATEMENTS, []),
        )),
        rescue(  # type: ignore
            lambda _: apply_default(cfg.get(DEFAULT)),
        ),
    )


def handle_attribute(
    collection: Union[Dict[str, Any], List[Any]],
    cfg: dict,
) -> ResultE[MapValue]:
    """Handle one attribute with mappings, ifs, casting and default value.

    :param collection: The collection of data to find data in
    :type collection: Union[Dict[str, Any], List[Any]]

    :param configuration: :term:`configuration` data to use when mapping
    :type configuration: Dict[str, Any]

    :return: Success/Failure containers
    :rtype: MapValue

    configuration expected to look like this:

    .. code-block:: json

        {
            "mappings": [],  # array of mapping objects
            "separator": None,
            "if_statements": [],  # array of if statement objects
            "casting": {}  # casting object, for casting types
            "default": "default value"
        }

    flow description:

    Map all objects in cfg[MAPPINGS] ->
    Apply separator to values if there are more than 1
    Failure -> fix to Success(None)
    Apply if statements
    Success -> Cast Value
    Failure -> apply default value

    Return Result
    """
    mapped_values = [
        mapped.unwrap()
        for mapped in
        [
            handle_mapping(collection, mapping)
            for mapping in cfg.get(MAPPINGS, [])
        ]
        if is_successful(mapped)
    ]

    # partially declare if statement and casting functions
    ifs = partial(apply_if_statements, if_objects=cfg.get(IF_STATEMENTS, []))
    cast = partial(apply_casting, casting=cfg.get(CASTING, {}))

    return flow(
        apply_separator(mapped_values, separator=cfg.get(SEPARATOR, '')),
        fix(lambda _: None),  # type: ignore
        bind(ifs),
        bind(cast),
        rescue(
            lambda _: apply_default(default=cfg.get(DEFAULT)),
        ),
    )



from attr import dataclass
from pycountry import countries, currencies, languages
from returns.result import safe
from typing_extensions import final

from piri.constants import ALPHA_THREE, ALPHA_TWO, NAME, NUMERIC, OFFICIAL_NAME


@final
@dataclass(frozen=True, slots=True)
class GetCountryCode(object):
    """Get Country code from string."""

    # dependiencies, country package
    _countries = countries
    _get = safe(countries.get)

    @safe
    def __call__(self, code: str) -> dict:
        """Find country by code or name."""
        code = str(code).strip()
        country = self._countries.get(alpha_2=code.upper())
        country = country or self._countries.get(alpha_3=code.upper())
        country = country or self._countries.get(numeric=code.zfill(3))
        country = country or self._countries.get(name=code)
        country = country or self._countries.get(official_name=code)

        if not country:
            raise ValueError(
                '{message}({country})'.format(
                    message='Could not find country matching value',
                    country=code,
                ),
            )

        return {
            ALPHA_TWO: country.alpha_2.upper(),
            ALPHA_THREE: country.alpha_3.upper(),
            NAME: country.name,
            NUMERIC: country.numeric,
            OFFICIAL_NAME: self._get_official_name(country).value_or(None),
        }

    @safe
    def _get_official_name(self, country) -> str:
        return country.official_name


@final
@dataclass(frozen=True, slots=True)
class GetCurrencyCode(object):
    """Get Currency code from string."""

    # dependiencies, country package
    _currencies = currencies
    _get = currencies.get

    @safe
    def __call__(self, code: str) -> dict:
        """Try to create a Currency object."""
        code = str(code).strip()

        cur = self._currencies.get(alpha_3=code.upper())
        cur = cur or self._currencies.get(numeric=code.zfill(3))
        cur = cur or self._currencies.get(name=code)

        if not cur:
            raise ValueError(
                '{message}({code})'.format(
                    message='Could not find currency matching code',
                    code=code,
                ),
            )

        return {
            ALPHA_THREE: cur.alpha_3.upper(),
            NUMERIC: cur.numeric,
            NAME: cur.name,
        }


@final
@dataclass(frozen=True, slots=True)
class GetLanguageCode(object):
    """SRO to get language code."""

    _languages = languages

    @safe
    def __call__(self, code: str) -> dict:
        """Try to create a Language object."""
        code = str(code).strip()
        lan = self._languages.get(alpha_2=code.lower())
        lan = lan or self._languages.get(alpha_3=code.lower())
        lan = lan or self._languages.get(name=code)

        if not lan:
            raise ValueError(
                '{message}({code})'.format(
                    message='Could not find language matching value',
                    code=code,
                ),
            )
        return {
            ALPHA_TWO: lan.alpha_2.upper(),
            ALPHA_THREE: lan.alpha_3.upper(),
            NAME: lan.name,
            'scope': lan.scope,
            'type': lan.type,
        }



from typing import Callable, Union

from attr import dataclass
from returns.functions import raise_exception
from returns.pipeline import is_successful
from returns.result import ResultE
from typing_extensions import final

from piri.mapper import map_data
from piri.schema import SchemaValidator


@final
@dataclass(frozen=True, slots=True)
class Process(object):
    """Process Callable Object."""

    validate: Callable[[dict], ResultE[dict]] = SchemaValidator()

    def __call__(
        self,
        input_data: dict,
        configuration: dict,
    ) -> ResultE[Union[list, dict]]:
        """Validate configuration then process data."""
        cfg = self.validate(configuration)

        if not is_successful(cfg):
            return cfg

        return map_data(input_data, cfg.unwrap())


def process(
    input_data: dict,
    configuration: dict,
) -> Union[list, dict]:
    """Call Process and unwrap value if no error, otherwise raise."""
    return Process()(
        input_data,
        configuration,
    ).alt(
        raise_exception,
    ).unwrap()
